import fcntl
import os
import yaml
import zookeeper

from twisted.internet.defer import inlineCallbacks, fail
from twisted.python.failure import Failure

from txzookeeper import ZookeeperClient
from txzookeeper.tests.utils import deleteTree

from juju.charm.publisher import CharmPublisher
from juju.charm.bundle import CharmBundle
from juju.charm.directory import CharmDirectory
from juju.providers.dummy import FileStorage
from juju.state.charm import CharmStateManager
from juju.state.errors import StateChanged
from juju.tests.common import get_test_zookeeper_address

from juju.environment.tests.test_config import (
    EnvironmentsConfigTestBase, SAMPLE_ENV)
from juju.lib.mocker import MATCH

from .test_repository import RepositoryTestBase


def _count_open_files():
    count = 0
    for sfd in os.listdir("/proc/self/fd"):
        ifd = int(sfd)
        if ifd < 3:
            continue
        try:
            fcntl.fcntl(ifd, fcntl.F_GETFD)
            count += 1
        except IOError:
            pass
    return count


class CharmPublisherTest(RepositoryTestBase):

    @inlineCallbacks
    def setUp(self):
        super(CharmPublisherTest, self).setUp()
        zookeeper.set_debug_level(0)

        self.charm = CharmDirectory(self.sample_dir1)
        # charm state manager, charm id
        self.charm_key = "local:%s-%s" % (
            self.charm.metadata.name, self.charm.metadata.revision)
        # provider storage key
        self.charm_storage_key = "%s:%s" % (
            self.charm_key, self.charm.get_sha256())

        self.client = ZookeeperClient(get_test_zookeeper_address())
        self.storage_dir = self.makeDir()
        self.storage = FileStorage(self.storage_dir)
        self.publisher = CharmPublisher(self.client, self.storage)

        yield self.client.connect()
        yield self.client.create("/charms")

    def tearDown(self):
        deleteTree("/", self.client.handle)
        self.client.close()
        super(CharmPublisherTest, self).tearDown()

    @inlineCallbacks
    def test_add_charm_and_publish(self):
        open_file_count = _count_open_files()
        yield self.publisher.add_charm("local", self.charm)
        result = yield self.publisher.publish()
        self.assertEquals(_count_open_files(), open_file_count)
        children = yield self.client.get_children("/charms")
        self.assertIn(self.charm_key, children)
        fh = yield self.storage.get("charms/%s" % self.charm_storage_key)
        bundle = CharmBundle(fh)
        self.assertEqual(self.charm.get_sha256(), bundle.get_sha256())

        self.assertEqual(
            result[0].bundle_url, "file://%s/charms/%s" % (
                self.storage_dir, self.charm_storage_key))

    @inlineCallbacks
    def test_published_charm_sans_unicode(self):
        yield self.publisher.add_charm("local", self.charm)
        yield self.publisher.publish()
        data, stat = yield self.client.get("/charms/%s" % self.charm_key)
        self.assertNotIn("unicode", data)

    @inlineCallbacks
    def test_add_charm_with_concurrent(self):
        """
        Publishing a charm, that has become published concurrent, after the
        add_charm, works fine. it will write to storage regardless. The use
        of a sha256 as part of the storage key is utilized to help ensure
        uniqueness of bits. The sha256 is also stored with the charm state.

        This relation betewen the charm state and the binary bits, helps
        guarantee the property that any published charm in zookeeper will use
        the binary bits that it was published with.
        """

        yield self.publisher.add_charm("local", self.charm)

        concurrent_publisher = CharmPublisher(
            self.client, self.storage)

        charm = CharmDirectory(self.sample_dir1)
        yield concurrent_publisher.add_charm("local", charm)

        yield self.publisher.publish()

        # modify the charm to create a conflict scenario
        self.makeFile("zebra",
                      path=os.path.join(self.sample_dir1, "junk.txt"))

        # assert the charm now has a different sha post modification
        modified_charm_sha = charm.get_sha256()
        self.assertNotEqual(
            modified_charm_sha,
            self.charm.get_sha256())

        # verify publishing raises a stateerror
        def verify_failure(result):
            if not isinstance(result, Failure):
                self.fail("Should have raised state error")
            result.trap(StateChanged)
            return True

        yield concurrent_publisher.publish().addBoth(verify_failure)

        # verify the zk state
        charm_nodes = yield self.client.get_children("/charms")
        self.assertEqual(charm_nodes, [self.charm_key])

        content, stat = yield self.client.get(
            "/charms/%s" % self.charm_key)

        # assert the checksum matches the initially published checksum
        self.assertEqual(
            yaml.load(content)["sha256"],
            self.charm_storage_key.split(":")[-1])

        stored_files = os.listdir(os.path.join(self.storage_dir, "charms"))
        self.assertIn(self.charm_storage_key, stored_files)

        # and the binary bits where stored
        modified_charm_storage_key = "local:%s-%s:%s" % (
            self.charm.metadata.name, self.charm.metadata.revision,
            modified_charm_sha)
        self.assertIn(modified_charm_storage_key, stored_files)

    @inlineCallbacks
    def test_add_charm_with_concurrent_removal(self):
        """
        If a charm is published, and it detects that the charm exists
        already exists, it will attempt to retrieve the charm state to
        verify there is no checksum mismatch. If concurrently the charm
        is removed, the publisher should fail with a statechange error.
        """
        charm_id = self.charm.metadata.get_charm_id("local")
        manager = self.mocker.patch(CharmStateManager)

        manager.get_charm_state(charm_id)
        self.mocker.passthrough()

        def match_charm_bundle(bundle):
            return isinstance(bundle, CharmBundle)

        def match_charm_url(url):
            return url.startswith("file://")

        manager.add_charm_state(
            "local", MATCH(match_charm_bundle), MATCH(match_charm_url))
        self.mocker.result(fail(zookeeper.NodeExistsException()))

        manager.get_charm_state(charm_id)
        self.mocker.result(fail(zookeeper.NoNodeException()))
        self.mocker.replay()

        yield self.publisher.add_charm("local", self.charm)
        yield self.failUnlessFailure(self.publisher.publish(), StateChanged)

    @inlineCallbacks
    def test_add_charm_already_known(self):
        """Adding an existing charm, is an effective noop, as its not added
        to the internal publisher queue.
        """
        # Do an initial publishing of the charm
        scheduled = yield self.publisher.add_charm("local", self.charm)
        self.assertTrue(scheduled)
        result = yield self.publisher.publish()
        self.assertEqual(result[0].name, self.charm.metadata.name)

        publisher = CharmPublisher(self.client, self.storage)
        scheduled = yield publisher.add_charm("local", self.charm)
        self.assertFalse(scheduled)
        scheduled = yield publisher.add_charm("local", self.charm)
        self.assertFalse(scheduled)

        result = yield publisher.publish()
        self.assertEqual(result[0].name, self.charm.metadata.name)
        self.assertEqual(result[1].name, self.charm.metadata.name)


class EnvironmentPublisherTest(EnvironmentsConfigTestBase):

    def setUp(self):
        super(EnvironmentPublisherTest, self).setUp()
        self.write_config(SAMPLE_ENV)
        self.config.load()
        self.environment = self.config.get("myfirstenv")
        zookeeper.set_debug_level(0)

    @inlineCallbacks
    def test_publisher_for_environment(self):
        publisher = yield CharmPublisher.for_environment(self.environment)
        self.assertTrue(isinstance(publisher, CharmPublisher))
