import os
import inspect
import shutil

import yaml
from twisted.internet.defer import inlineCallbacks

from juju.lib.testing import TestCase
from juju.charm.directory import CharmDirectory
from juju.charm.errors import CharmNotFound
from juju.charm.repository import LocalCharmRepository

from juju.charm import tests


unbundled_repository = os.path.join(
    os.path.dirname(inspect.getabsfile(tests)), "repository")


class RepositoryTestBase(TestCase):

    @inlineCallbacks
    def setUp(self):
        yield super(RepositoryTestBase, self).setUp()
        self.bundled_repo_path = self.makeDir()
        self.unbundled_repo_path = self.makeDir()

        os.rmdir(self.unbundled_repo_path)

        shutil.copytree(unbundled_repository, self.unbundled_repo_path)

        self.sample_dir1 = os.path.join(self.unbundled_repo_path, "old")
        self.sample_dir2 = os.path.join(self.unbundled_repo_path, "new")


class RepositoryTest(RepositoryTestBase):

    def setUp(self):
        super(RepositoryTest, self).setUp()

        # bundle sample charms
        CharmDirectory(self.sample_dir1).make_archive(
            os.path.join(self.bundled_repo_path, "old.frml"))
        CharmDirectory(self.sample_dir2).make_archive(
            os.path.join(self.bundled_repo_path, "new.frml"))

        # define repository objects
        self.repository1 = LocalCharmRepository(self.unbundled_repo_path)
        self.repository2 = LocalCharmRepository(self.bundled_repo_path)

    def test_find_charm_by_name_which_is_unbundled(self):
        charm = self.repository1.find("sample")
        self.assertTrue(charm)
        self.assertTrue(charm.metadata)

    def test_find_charm_ignores_unknown_files(self):
        self.makeFile(
            "Foobar",
            path=os.path.join(self.repository1.path, "zebra"))
        self.assertRaises(CharmNotFound, self.repository1.find, "zebra")

    def test_find_charm_ignores_unknown_directories(self):
        self.makeDir(
            path=os.path.join(self.repository1.path, "zebra"))
        self.assertRaises(CharmNotFound, self.repository1.find, "zebra")

    def test_find_charm_ignores_broken_charms(self):
        charm_path = self.makeDir(
            path=os.path.join(self.repository1.path, "zebra"))
        fh = open(os.path.join(charm_path, "metadata.yaml"), "w+")
        fh.write("""\
        description: helo
juju: charm
name: hi
requires: {interface: zebra
revision: 0
summary: hola""")
        fh.close()

        self.assertRaises(CharmNotFound, self.repository1.find, "zebra")

    def test_find_charm_by_name_which_is_bundled(self):
        charm = self.repository2.find("sample")
        self.assertTrue(charm)
        self.assertTrue(charm.metadata)

    def test_find_charm_by_name_fails(self):
        error = self.assertRaises(
            CharmNotFound, self.repository1.find, "missing")
        self.assertEquals(str(error),
                          "Charm 'missing' not found in repository %s"
                          % self.unbundled_repo_path)

    def test_find_charm_with_multiple_versions_returns_latest(self):
        # Copy the repository out of the codebase, so that we can hack it.
        charm = self.repository1.find("sample")
        self.assertEquals(charm.metadata.revision, 2)

        # Invert the "latest" logic, to ensure it's not just a coincidence.
        file = open(
            os.path.join(self.repository1.path, "old/metadata.yaml"), "rw+")
        data = yaml.load(file.read())
        data["revision"] = 3
        file.seek(0)
        file.write(yaml.dump(data))
        file.close()

        charm = self.repository1.find("sample")
        self.assertEquals(charm.metadata.revision, 3)
