"""Implementation of add unit subcommand"""

from twisted.internet.defer import inlineCallbacks

from juju.control.utils import get_environment
from juju.state.machine import MachineStateManager
from juju.state.placement import place_unit, UNASSIGNED_POLICY
from juju.state.service import ServiceStateManager


def configure_subparser(subparsers):
    """Configure add-unit subcommand"""
    sub_parser = subparsers.add_parser("add-unit", help=command.__doc__)
    sub_parser.add_argument(
        "--environment", "-e",
        help="juju environment to operate in.")
    sub_parser.add_argument(
        "service_name",
        help="Name of the service a unit should be added for")
    sub_parser.add_argument(
        "--placement",
        help="Placement policy used to place units of this service")

    return sub_parser


def command(options):
    """Add a new service unit."""
    environment = get_environment(options)
    env_defaults = environment.get_serialization_data()
    if not options.placement:
        options.placement = env_defaults[environment.name].get(
            "placement", UNASSIGNED_POLICY)
    return add_unit(
        options.environments,
        environment,
        options.verbose,
        options.log,
        options.service_name,
        options.placement)


@inlineCallbacks
def add_unit(config, environment, verbose, log, service_name, policy_name):
    provider = environment.get_machine_provider()
    client = yield provider.connect()
    try:
        service_manager = ServiceStateManager(client)
        machine_manager = MachineStateManager(client)

        service_state = yield service_manager.get_service_state(service_name)
        unit_state = yield service_state.add_unit_state()

        yield place_unit(client, policy_name, machine_manager, unit_state)
        log.info("Unit %r added to service %r",
            unit_state.unit_name, service_state.service_name)
    finally:
        yield client.close()
