"""Implementation of charm-upgrade subcommand"""

import os

from twisted.internet.defer import inlineCallbacks

from juju.control.utils import get_environment

from juju.charm.errors import NewerCharmNotFound
from juju.charm.metadata import parse_charm_id
from juju.charm.publisher import CharmPublisher
from juju.charm.repository import LocalCharmRepository

from juju.state.service import ServiceStateManager
from juju.unit.workflow import is_unit_running


def configure_subparser(subparsers):
    """Configure charm-upgrade subcommand"""
    sub_parser = subparsers.add_parser("upgrade-charm", help=command.__doc__)
    sub_parser.add_argument(
        "--dry-run", "-n", action="store_true",
        help="Dry-Run, show which charm would be deployed for upgrade.")
    sub_parser.add_argument(
        "--environment", "-e",
        help="juju environment to operate in.")
    sub_parser.add_argument(
        "--repository", required=True,
        help="Directory for charm lookup and retrieval",
        type=os.path.abspath)
    sub_parser.add_argument(
        "service_name",
        help="Name of the service that should be upgraded")
    return sub_parser


def command(options):
    """Upgrade a service's charm."""
    environment = get_environment(options)
    return upgrade_charm(
        options.environments,
        environment,
        options.verbose,
        options.log,
        options.repository,
        options.service_name,
        options.dry_run)


@inlineCallbacks
def upgrade_charm(
    config, environment, verbose, log, repository_path, service_name, dry_run):
    """Upgrades a service's charm.

    First determines if an upgrade is available, then updates the
    service charm reference, and marks the units as needing upgrades.
    """
    provider = environment.get_machine_provider()
    client = yield provider.connect()

    repository = LocalCharmRepository(repository_path)
    service_manager = ServiceStateManager(client)
    service_state = yield service_manager.get_service_state(service_name)
    charm_id = yield service_state.get_charm_id()

    namespace, name, current_revision = parse_charm_id(charm_id)

    # Lookup the charm in the repository
    charm = repository.find(name)

    # Verify its newer than what's deployed
    if not charm.metadata.revision > current_revision:
        if dry_run:
            log.info("Service already running latest charm %r", charm_id)
        else:
            raise NewerCharmNotFound(charm_id)
    elif dry_run:
        log.info("Service would be upgraded from charm %r to %r",
                 charm_id, charm.metadata.get_charm_id("local"))

    # On dry run, stop before modifying state.
    if not dry_run:
        # Publish the new charm
        storage = provider.get_file_storage()
        publisher = CharmPublisher(client, storage)
        yield publisher.add_charm("local", charm)
        result = yield publisher.publish()
        charm_state = result[0]

        # Update the service charm reference
        yield service_state.set_charm_id(charm_state.id)

    # Mark the units for upgrades
    units = yield service_state.get_all_unit_states()
    for unit in units:
        running, state = yield is_unit_running(client, unit)
        if not running:
            log.info(
                "Unit %r is not in a running state (state: %r), won't upgrade",
                unit.unit_name, state or "uninitialized")
            continue

        if not dry_run:
            yield unit.set_upgrade_flag()
