import os
import pipes
import subprocess
import sys
import tempfile

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.threads import deferToThread

from juju.errors import JujuError

DATA_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "data"))


class LXCError(JujuError):
    """Indicates a low level error with an LXC container"""


def _cmd(args):
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout_data, _ = p.communicate()
    r = p.returncode
    if  r != 0:
        # read the stdout/err streams and show the user
        print >>sys.stderr, stdout_data
        raise LXCError(stdout_data)
    return (r, stdout_data)


# Wrapped lxc cli primitives
def _lxc_create(
    container_name, template="ubuntu", config_file=None, release="oneiric"):
    # the -- argument indicates the last parameters are passed
    # to the template and not handled by lxc-create
    args = ["sudo", "lxc-create",
            "-n", container_name,
            "-t", template,
            "-f", config_file,
            "--",
            "-r", release]
    return _cmd(args)


def _lxc_start(container_name, console_log=None):
    args = ["sudo", "lxc-start", "--daemon", "-n", container_name]
    if console_log:
        args.append("-c")
        args.append(console_log)
    return _cmd(args)


def _lxc_stop(container_name):
    _cmd(["sudo", "lxc-stop", "-n", container_name])


def _lxc_destroy(container_name):
    return _cmd(["sudo", "lxc-destroy", "-n", container_name])


def _lxc_ls():
    _, output = _cmd(["lxc-ls"])
    output = output.replace("\n", " ")
    return set([c for c in output.split(" ") if c])


def _lxc_wait(container_name, state="RUNNING"):
    """Wait for container to be in a given state RUNNING|STOPPED."""

    def wait(container_name):
        rc, _ = _cmd(["sudo", "lxc-wait",
                   "-n", container_name,
                   "-s", state])
        return rc == 0

    return deferToThread(wait, container_name)


def _customize_container(customize_script, container_root):
    if not os.path.isdir(container_root):
        raise LXCError("Expect container root directory: %s" %
                       container_root)
    # write the container scripts into the container
    fd, in_path = tempfile.mkstemp(prefix=os.path.basename(customize_script),
                                   dir=os.path.join(container_root, "tmp"))

    os.write(fd, open(customize_script, "r").read())
    os.close(fd)
    os.chmod(in_path, 0755)

    args = ["sudo", "chroot", container_root,
            os.path.join("/tmp", os.path.basename(in_path))]
    return _cmd(args)


def validate_path(pathname):
    if not os.access(pathname, os.R_OK):
        raise LXCError("Invalid or unreadable file: %s" % pathname)


@inlineCallbacks
def get_containers(prefix):
    """Return a dictionary of containers key names to runtime boolean value.

    :param prefix: Optionally specify a prefix that the container should
    match any returned containers.
    """
    _, output = yield deferToThread(_cmd, ["lxc-ls"])

    containers = {}
    for i in filter(None, output.split("\n")):
        if i in containers:
            containers[i] = True
        else:
            containers[i] = False

    if prefix:
        remove = [k for k in containers.keys() if not
                  k.startswith(prefix)]
        map(containers.pop, remove)

    returnValue(containers)


class LXCContainer(object):
    def __init__(self,
                 container_name, configuration=None, network_name="virbr0",
                 customize_script=None):
        """Create an LXCContainer

        :param container_name: should be unique within the system

        :param configuration: dict of values which are written into
        the containers /etc/juju/juju.conf file and expected
        to be consumed by the customize script

        :param network_name: name of network link

        :param customize_script: script used inside container to make
        it juju ready
        """

        self.container_name = container_name

        if customize_script is None:
            customize_script = os.path.join(DATA_PATH,
                                            "juju-create")
        self.customize_script = customize_script
        validate_path(self.customize_script)

        self.configuration = configuration
        self.network_name = network_name
        self.running = False

    @property
    def rootfs(self):
        return "/var/lib/lxc/%s/rootfs/" % self.container_name

    def _p(self, path):
        if path[0] == "/":
            path = path[1:]
        return os.path.join(self.rootfs, path)

    def _make_lxc_config(self, network_name):
        lxc_config = os.path.join(DATA_PATH, "lxc.conf")
        with open(lxc_config, "r") as fh:
            template = fh.read()
            fd, output_fn = tempfile.mkstemp(suffix=".conf")
            output_config = open(output_fn, "w")
            output_config.write(template % {"network_name": network_name})
            output_config.close()

            validate_path(output_fn)
            return output_fn

    @inlineCallbacks
    def _customize_container(self):
        config_dir = self._p("etc/juju")
        if not os.path.exists(config_dir):
            _cmd(["sudo", "mkdir", config_dir])

        if not self.configuration:
            return

        _, fn = tempfile.mkstemp()
        with open(fn, "w") as fp:
            for name, value in self.configuration.items():
                fp.write("%s=%s\n" % (name.upper(), pipes.quote(value)))

        _cmd(["sudo", "mv", fn,
              self._p("etc/juju/juju.conf")])

        yield deferToThread(
            _customize_container, self.customize_script, self.rootfs)

    @inlineCallbacks
    def execute(self, args):
        if not isinstance(args, (list, tuple)):
            args = [args, ]

        args = ["sudo", "chroot", self.rootfs] + args
        yield deferToThread(_cmd, args)

    @inlineCallbacks
    def create(self):
        # open the template file and create a new temp processed
        # version
        lxc_config = self._make_lxc_config(self.network_name)
        yield deferToThread(_lxc_create, self.container_name,
                            config_file=lxc_config)
        yield self._customize_container()
        os.unlink(lxc_config)

    @inlineCallbacks
    def run(self):
        if not os.path.exists(self.rootfs):
            yield self.create()

        yield deferToThread(_lxc_start, self.container_name)
        self.running = yield _lxc_wait(self.container_name, "RUNNING")

    @inlineCallbacks
    def stop(self):
        yield deferToThread(_lxc_stop, self.container_name)
        yield _lxc_wait(self.container_name, "STOPPED")
        self.running = False

    @inlineCallbacks
    def destroy(self):
        yield self.stop()
        yield deferToThread(_lxc_destroy, self.container_name)
