import os
import tempfile

from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import deferToThread

from juju.lib.lxc import (_lxc_start, _lxc_stop, _lxc_create,
                              _lxc_wait, _lxc_ls, _lxc_destroy,
                              LXCContainer, get_containers)
from juju.lib.testing import TestCase, get_test_zookeeper_address


def run_lxc_tests():
    if os.environ.get("TEST_LXC"):
        return None
    return "TEST_LXC=1 to include lxc tests"


DATA_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "data"))


DEFAULT_CONTAINER = "lxc_test"


class LXCTest(TestCase):
    timeout = 240
    skip = run_lxc_tests()

    def setUp(self):
        self.config = self.make_config()

        @self.addCleanup
        def remove_config():
            if os.path.exists(self.config):
                os.unlink(self.config)

    def make_config(self, network_name="virbr0"):
        lxc_config = os.path.join(DATA_PATH, "lxc.conf")
        template = open(lxc_config, "r").read()

        fd, output_fn = tempfile.mkstemp(suffix=".conf")
        output_config = open(output_fn, "w")
        output_config.write(template % {"network_name": network_name})
        output_config.close()

        return output_fn

    def cleanContainer(self, container_name):
        if os.path.exists("/var/lib/lxc/%s" % container_name):
            _lxc_stop(container_name)
            _lxc_destroy(container_name)

    def test_lxc_create(self):
        self.addCleanup(self.cleanContainer, DEFAULT_CONTAINER)

        _lxc_create(DEFAULT_CONTAINER, config_file=self.config)

        # verify we can find the container
        output = _lxc_ls()
        self.assertIn(DEFAULT_CONTAINER, output)

        # remove and verify the container was removed
        _lxc_destroy(DEFAULT_CONTAINER)
        output = _lxc_ls()
        self.assertNotIn(DEFAULT_CONTAINER, output)

    def test_lxc_start(self):
        self.addCleanup(self.cleanContainer, DEFAULT_CONTAINER)

        _lxc_create(DEFAULT_CONTAINER, config_file=self.config)

        _lxc_start(DEFAULT_CONTAINER)
        _lxc_stop(DEFAULT_CONTAINER)

    @inlineCallbacks
    def test_lxc_deferred(self):
        self.addCleanup(self.cleanContainer, DEFAULT_CONTAINER)
        yield deferToThread(_lxc_create, DEFAULT_CONTAINER, config_file=self.config)
        yield deferToThread(_lxc_start, DEFAULT_CONTAINER)

    @inlineCallbacks
    def test_lxc_container(self):
        self.addCleanup(self.cleanContainer, DEFAULT_CONTAINER)

        c = LXCContainer(DEFAULT_CONTAINER,
                         dict(juju_zookeeper=get_test_zookeeper_address(),
                              juju_origin="distro"))
        self.assertFalse(c.running)
        yield c.run()
        self.assertTrue(c.running)

        output = _lxc_ls()
        self.assertIn(DEFAULT_CONTAINER, output)

        # verify we have a path into the container
        self.assertTrue(os.path.exists(c.rootfs))

        # expect a processed config file above the rootfs
        # -- this requires root to read
        #network = open(os.path.join(c.rootfs, "..", "config"), "r").read()
        #self.assertIn("lxc.network.link=virbr0", network)

        # check that some of the customize modifications were made
        self.assertTrue(os.path.exists(os.path.join(
                    c.rootfs, "etc", "juju")))
        self.assertTrue(os.path.exists(os.path.join(
                    c.rootfs, "usr", "bin", "juju")))

        # verify that we have shell variables in rootfs/etc/juju.conf
        config = open(os.path.join(c.rootfs, "etc", "juju", "juju.conf"), "r").read()
        self.assertIn("JUJU_ZOOKEEPER=%s" % get_test_zookeeper_address(), config)

        # verify that we are in containers
        containers = yield get_containers(None)
        self.assertEqual(containers[DEFAULT_CONTAINER], True)

        # tear it down
        yield c.destroy()
        self.assertFalse(c.running)

        containers = yield get_containers(None)
        self.assertNotIn(DEFAULT_CONTAINER, containers)

        # and its gone
        output = _lxc_ls()
        self.assertNotIn(DEFAULT_CONTAINER, output)

    @inlineCallbacks
    def test_lxc_wait(self):
        self.addCleanup(self.cleanContainer, DEFAULT_CONTAINER)

        _lxc_create(DEFAULT_CONTAINER, config_file=self.config)

        _lxc_start(DEFAULT_CONTAINER)

        def waitForState(result):
            self.assertEqual(result, True)

        d = _lxc_wait(DEFAULT_CONTAINER, "RUNNING")
        d.addCallback(waitForState)
        yield d

        _lxc_stop(DEFAULT_CONTAINER)
        yield _lxc_wait(DEFAULT_CONTAINER, "STOPPED")
        _lxc_destroy(DEFAULT_CONTAINER)





