"""
Service Unit Deployment unit tests
"""
import os
import sys

from twisted.internet.protocol import ProcessProtocol
from twisted.internet.defer import inlineCallbacks

import juju

from juju.charm import get_charm_from_path
from juju.charm.tests.test_repository import RepositoryTestBase
from juju.lib.mocker import MATCH
from juju.lib.twistutils import get_module_directory
from juju.machine.unit import ServiceUnitDeployment
from juju.machine.errors import UnitDeploymentError
from juju.tests.common import get_test_zookeeper_address


MATCH_PROTOCOL = MATCH(lambda x: isinstance(x, ProcessProtocol))


class ServiceUnitDeploymentTest(RepositoryTestBase):

    def setUp(self):
        super(ServiceUnitDeploymentTest, self).setUp()
        self.charm = get_charm_from_path(self.sample_dir1)
        self.bundle = self.charm.as_bundle()
        self.juju_directory = self.makeDir()
        self.units_directory = self.makeDir(dirname=self.juju_directory)
        self.unit_name = "wordpress/0"

        self.deployment = ServiceUnitDeployment(
            self.unit_name,
            self.units_directory)

        self.assertEqual(
            self.deployment.unit_agent_module, "juju.agents.unit")
        self.deployment.unit_agent_module = "juju.agents.dummy"

    def process_kill(self, pid):
        try:
            os.kill(pid, 9)
        except OSError:
            pass

    def test_unit_name_with_path_manipulation_raises_assertion(self):
        self.assertRaises(
            AssertionError,
            ServiceUnitDeployment,
            "../../etc/password/zebra/0",
            self.units_directory)

    def test_unit_directory(self):
        self.assertEqual(
            self.deployment.directory,
            os.path.join(self.units_directory,
                         self.unit_name.replace("/", "-")))

    def test_unit_pid_file(self):
        self.assertEqual(
            self.deployment.pid_file,
            os.path.join(self.units_directory,
                         "%s.pid" % (self.unit_name.replace("/", "-"))))

    def test_service_unit_start(self):
        """
        Starting a service unit will result in a unit workspace being created
        if it does not exist and a running service unit agent.
        """
        self.deployment.unpack_charm(self.bundle)
        d = self.deployment.start("0", get_test_zookeeper_address())

        @inlineCallbacks
        def validate_result(result):
            # give process time to write its pid
            yield self.sleep(0.1)
            self.addCleanup(
                self.process_kill,
                int(open(self.deployment.pid_file).read()))
            self.assertEqual(result, True)

        d.addCallback(validate_result)
        return d

    def test_service_unit_start_with_integer_machine_id(self):
        """
        Starting a service unit will result in a unit workspace being created
        if it does not exist and a running service unit agent.
        """
        self.deployment.unpack_charm(self.bundle)
        d = self.deployment.start(21, get_test_zookeeper_address())

        @inlineCallbacks
        def validate_result(result):
            # give process time to write its pid
            yield self.sleep(0.1)
            self.addCleanup(
                self.process_kill,
                int(open(self.deployment.pid_file).read()))
            self.assertEqual(result, True)

        d.addCallback(validate_result)
        return d

    def test_service_unit_start_with_agent_startup_error(self):
        """
        Starting a service unit will result in a unit workspace being created
        if it does not exist and a running service unit agent.
        """
        self.deployment.unit_agent_module = "magichat.xr1"
        self.deployment.unpack_charm(self.bundle)
        d = self.deployment.start("0", get_test_zookeeper_address())

        self.failUnlessFailure(d, UnitDeploymentError)

        def validate_result(error):
            self.assertIn("No module named magichat", str(error))

        d.addCallback(validate_result)
        return d

    def test_service_unit_start_agent_arguments(self):
        """
        Starting a service unit will start a service unit agent with arguments
        denoting the current machine id, zookeeper server location, and the
        unit name. Additionally it will configure the log and pid file
        locations.
        """
        machine_id = "0"
        zookeeper_hosts = "menagerie.example.com:2181"

        from twisted.internet import reactor
        mock_reactor = self.mocker.patch(reactor)

        environ = dict(os.environ)
        environ["JUJU_UNIT_NAME"] = self.unit_name
        environ["JUJU_HOME"] = self.juju_directory
        environ["JUJU_MACHINE_ID"] = machine_id
        environ["JUJU_ZOOKEEPER"] = zookeeper_hosts
        environ["PYTHONPATH"] = ":".join(
            filter(None, [
                os.path.dirname(get_module_directory(juju)),
                environ.get("PYTHONPATH")]))

        pid_file = os.path.join(
            self.units_directory,
            "%s.pid" % self.unit_name.replace("/", "-"))

        log_file = os.path.join(
            self.deployment.directory,
            "charm.log")

        args = [sys.executable, "-m", "juju.agents.dummy", "-n",
                "--pidfile", pid_file, "--logfile", log_file]

        mock_reactor.spawnProcess(
            MATCH_PROTOCOL, sys.executable, args, environ)

        self.mocker.replay()
        self.deployment.unpack_charm(self.bundle)
        self.deployment.start(machine_id, zookeeper_hosts)

    def test_service_unit_start_pre_unpack(self):
        """
        Attempting to start a charm before the charm is unpacked
        results in an exception.
        """
        error = self.assertRaises(
            UnitDeploymentError,
            self.deployment.start, "0", get_test_zookeeper_address())
        self.assertEquals(str(error), "Charm must be unpacked first.")

    @inlineCallbacks
    def test_service_unit_destroy(self):
        """
        Forcibly stop a unit, and destroy any directories associated to it
        on the machine, and kills the unit agent process.
        """
        self.deployment.unpack_charm(self.bundle)
        yield self.deployment.start("0", get_test_zookeeper_address())
        # give the process time to write its pid file
        yield self.sleep(0.1)
        pid = int(open(self.deployment.pid_file).read())
        self.deployment.destroy()
        # give the process time to die.
        yield self.sleep(0.1)
        e = self.assertRaises(OSError, os.kill, pid, 0)
        self.assertEqual(e.errno, 3)
        self.assertFalse(os.path.exists(self.deployment.directory))
        self.assertFalse(os.path.exists(self.deployment.pid_file))

    def test_service_unit_destroy_stale_pid(self):
        """
        A stale pid file does not cause any errors.

        We mock away is_running as otherwise it will check for this, but
        there exists a small window when the result may disagree.
        """
        self.makeFile("8917238", path=self.deployment.pid_file)
        mock_deployment = self.mocker.patch(self.deployment)
        mock_deployment.is_running()
        self.mocker.result(True)
        self.mocker.replay()
        self.deployment.destroy()

    def test_service_unit_destroy_perm_error(self):
        """
        A stale pid file does not cause any errors.

        We mock away is_running as otherwise it will check for this, but
        there exists a small window when the result may disagree.
        """
        if os.geteuid() == 0:
            return
        self.makeFile("1", path=self.deployment.pid_file)
        mock_deployment = self.mocker.patch(self.deployment)
        mock_deployment.is_running()
        self.mocker.result(True)
        self.mocker.replay()
        self.assertRaises(OSError,
                          self.deployment.destroy)

    def test_service_unit_destroy_undeployed(self):
        """
        If the unit is has not been deployed, nothing happens.
        """
        self.deployment.destroy()
        self.assertFalse(os.path.exists(self.deployment.directory))

    def test_service_unit_destroy_not_running(self):
        """
        If the unit is not running, then destroy will just remove
        its directory.
        """
        self.deployment.unpack_charm(self.bundle)
        self.assertTrue(os.path.exists(self.deployment.directory))
        self.deployment.destroy()
        self.assertFalse(os.path.exists(self.deployment.directory))

    def test_unpack_charm(self):
        """
        The deployment unpacks a charm bundle into the unit workspace.
        """
        self.deployment.unpack_charm(self.bundle)
        unit_path = os.path.join(
            self.units_directory, self.unit_name.replace("/", "-"))

        self.assertTrue(os.path.exists(unit_path))
        charm_path = os.path.join(unit_path, "charm")
        self.assertTrue(os.path.exists(charm_path))
        charm = get_charm_from_path(charm_path)
        self.assertEqual(
            charm.metadata.get_charm_id("local"),
            self.charm.metadata.get_charm_id("local"))

    def xtest_unpack_charm_exception_insufficient_disk_space(self):
        """If a charm can't be unpacked a deployment specific error is raised
        """
        self.assertRaises(
            self.deployment.unpack_charm(self.bundle))

    def test_unpack_charm_exception_invalid_charm(self):
        """
        If the charm bundle is corrupted or invalid a deployment specific
        error is raised.
        """
        error = self.assertRaises(
            UnitDeploymentError,
            self.deployment.unpack_charm, self.charm)
        self.assertEquals(
            str(error),
            "Invalid charm for deployment: %s" % self.charm.path)

    def test_is_running_no_pid_file(self):
        """
        If there is no pid file the service unit is not running.
        """
        self.assertEqual(self.deployment.is_running(), False)

    def test_is_running(self):
        """
        The service deployment will check the pid and validate
        that the pid found is a running process.
        """
        self.makeFile(
            str(os.getpid()), path=self.deployment.pid_file)
        self.assertEqual(self.deployment.is_running(), True)

    def test_is_running_against_unknown_error(self):
        """
        If we don't have permission to access the process, the
        original error should get passed along.
        """
        if os.geteuid() == 0:
            return
        self.makeFile("1", path=self.deployment.pid_file)
        self.assertRaises(OSError,
                          self.deployment.is_running)

    def test_is_running_invalid_pid_file(self):
        """
        If the pid file is corrupted on disk, and does not contain
        a valid integer, then the agent is not running.
        """
        self.makeFile("abcdef", path=self.deployment.pid_file)
        self.assertEqual(self.deployment.is_running(), False)

    def test_is_running_invalid_pid(self):
        """
        If the pid file refers to an invalid process then the
        agent is not running.
        """
        self.makeFile("669966", path=self.deployment.pid_file)
        self.assertEqual(self.deployment.is_running(), False)
