import os

from juju.errors import CloudInitError
from juju.lib.testing import TestCase
from juju.providers.common.cloudinit import CloudInit
from juju.providers.dummy import DummyMachine

DATA_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data")


class CloudInitTest(TestCase):

    def construct_normal(self):
        cloud_init = CloudInit()
        cloud_init.add_ssh_key("chubb")
        cloud_init.set_machine_id("passport")
        cloud_init.set_provider_type("dummy")
        cloud_init.set_zookeeper_machines([
            DummyMachine("blah", "blah", "cotswold"),
            DummyMachine("blah", "blah", "longleat")])
        return cloud_init

    def construct_bootstrap(self, with_zookeepers=False):
        cloud_init = CloudInit()
        cloud_init.enable_bootstrap()
        cloud_init.add_ssh_key("chubb")
        cloud_init.set_machine_id("passport")
        cloud_init.set_provider_type("dummy")
        cloud_init.set_instance_id_accessor("token")
        cloud_init.set_zookeeper_secret("seekrit")
        if with_zookeepers:
            cloud_init.set_zookeeper_machines([
                DummyMachine("blah", "blah", "cotswold"),
                DummyMachine("blah", "blah", "longleat")])
        return cloud_init

    def assert_render(self, cloud_init, name):
        with open(os.path.join(DATA_DIR, name)) as f:
            expected = f.read()
        self.assertEquals(cloud_init.render(), expected)

    def test_render_validate_normal(self):
        cloud_init = CloudInit()
        error = self.assertRaises(CloudInitError, cloud_init.render)
        self.assertEquals(
            str(error),
            "Incomplete cloud-init: you need to call add_ssh_key, "
            "set_machine_id, set_provider_type, set_zookeeper_machines")

    def test_render_validate_bootstrap(self):
        cloud_init = CloudInit()
        cloud_init.enable_bootstrap()
        error = self.assertRaises(CloudInitError, cloud_init.render)
        self.assertEquals(
            str(error),
            "Incomplete cloud-init: you need to call add_ssh_key, "
            "set_machine_id, set_provider_type, set_instance_id_accessor, "
            "set_zookeeper_secret")

    def test_source_validate(self):
        bad_choices = (
            (None,      False, False),
            (None,      True,  True),
            ("lp:blah", True,  True),
            ("lp:blah", False, True),
            ("lp:blah", True,  False))
        cloud_init = CloudInit()
        for choice in bad_choices:
            error = self.assertRaises(
                CloudInitError, cloud_init.set_juju_source, *choice)
            self.assertEquals(str(error), "Please specify one source")

    def test_render_normal(self):
        self.assert_render(self.construct_normal(), "cloud_init_normal")

    def test_render_ppa_source(self):
        cloud_init = self.construct_normal()
        cloud_init.set_juju_source(ppa=True)
        self.assert_render(cloud_init, "cloud_init_normal")

    def test_render_distro_source(self):
        cloud_init = self.construct_normal()
        cloud_init.set_juju_source(distro=True)
        self.assert_render(cloud_init, "cloud_init_distro")

    def test_render_branch_source(self):
        cloud_init = self.construct_normal()
        cloud_init.set_juju_source(branch="lp:blah/juju/blah-blah")
        self.assert_render(cloud_init, "cloud_init_branch")

    def test_render_bootstrap(self):
        self.assert_render(self.construct_bootstrap(), "cloud_init_bootstrap")

    def test_render_bootstrap_with_zookeepers(self):
        self.assert_render(
            self.construct_bootstrap(True), "cloud_init_bootstrap_zookeepers")
