from base64 import b64encode
from cStringIO import StringIO
from gzip import GzipFile

from twisted.internet.defer import inlineCallbacks, returnValue

from juju.providers.common.launch import LaunchMachine

from .machine import machine_from_dict

_LATE_COMMAND_TEMPLATE = """
seed_d=/var/lib/cloud/seed/nocloud-net
mkdir -p "$seed_d"
cat > "$seed_d"/meta-data <<"EOF"
instance-id: %s
EOF

cat > "$seed_d"/user-data <<"EOF"
%s
EOF
"""

_KSMETA_LATE_COMMAND_TEMPLATE = (
    "in-target sh -c '"
        "f=$1; "
        "shift; "
        "echo $0 | base64 --decode | gunzip > $f "
        "&& chmod u+x $f "
        "&& $f $*"
    "' %s /root/late-command")


def _base64_gzip(content):
    gzipped = StringIO()
    gzip_file = GzipFile(fileobj=gzipped, mode="wb", compresslevel=9)
    gzip_file.write(content)
    gzip_file.close()
    return b64encode(gzipped.getvalue())


def _ks_meta(**kwargs):
    return " ".join('%s="%s"' % item for item in kwargs.items())


class OrchestraLaunchMachine(LaunchMachine):
    """rchestra operation for launching an instance"""

    @inlineCallbacks
    def start_machine(self, machine_id, zookeepers):
        """Actually launch an instance with Orchestra.

        :param str machine_id: the juju machine ID to assign

        :param zookeepers: the machines currently running zookeeper, to which
            the new machine will need to connect
        :type zookeepers: list of
            :class:`juju.providers.orchestra.machine.OrchestraMachine`

        :return: a singe-entry list containing a
            :class:`juju.providers.orchestra.machine.OrchestraMachine`
            representing the newly-launched machine
        :rtype: :class:`twisted.internet.defer.Deferred`
        """
        cobbler = self._provider.cobbler
        instance_id = yield cobbler.acquire_system()

        cloud_init = self._create_cloud_init(machine_id, zookeepers)
        cloud_init.set_provider_type("orchestra")
        cloud_init.set_instance_id_accessor(instance_id)
        ks_meta = self._build_ks_meta(
            cloud_init.render(), machine_id, instance_id)

        yield cobbler.set_on_system(instance_id, "ks_meta", ks_meta)
        yield cobbler.start_system(instance_id)
        (info,) = yield cobbler.describe_systems(instance_id)
        returnValue([machine_from_dict(info)])

    def _build_ks_meta(self, user_data, machine_id, instance_id):
        late_command = _LATE_COMMAND_TEMPLATE % (instance_id, user_data)
        encoded = _base64_gzip(late_command)
        juju_late_command = _KSMETA_LATE_COMMAND_TEMPLATE % encoded
        return _ks_meta(
            MACHINE_ID=machine_id, JUJU_LATE_COMMAND=juju_late_command)
