import yaml

from twisted.internet.defer import (
    inlineCallbacks, returnValue, succeed)

from zookeeper import NoNodeException

from juju.charm.config import ConfigOptions
from juju.charm.metadata import MetaData
from juju.state.base import StateBase
from juju.state.errors import CharmStateNotFound


class CharmStateManager(StateBase):
    """Manages the state of charms in an environment."""

    @inlineCallbacks
    def add_charm_state(self, namespace, charm, url):
        """Register metadata about the provided Charm.

        :param namespace: The namespace this charm will be
            registered under.

        :param url: The provider storage url for a charm.
        """
        name = charm.metadata.name
        revision = charm.metadata.revision

        charm_data = {
            "config": charm.config.get_serialization_data(),
            "metadata": charm.metadata.get_serialization_data(),
            "sha256": charm.get_sha256(),
            "url": url
        }

        charm_id = "%s:%s-%s" % (namespace, name, revision)

        # XXX In the future we'll have to think about charm
        #     replacements here. For now this will do, and will
        #     explode reliably in case of conflicts.
        yield self._client.create("/charms/%s" % charm_id,
                                  yaml.safe_dump(charm_data))

        charm_state = CharmState(
            self._client, namespace, name, revision, charm_data)

        returnValue(charm_state)

    @inlineCallbacks
    def get_charm_state(self, charm_id):
        """Retrieve a CharmState for the given charm id."""
        try:
            content, stat = yield self._client.get("/charms/" + charm_id)
        except NoNodeException:
            raise CharmStateNotFound(charm_id)
        charm_data = yaml.load(content)
        rest, revision = charm_id.rsplit("-", 1)
        revision = int(revision)
        namespace, name = rest.split(":", 1)
        charm_state = CharmState(
            self._client, namespace, name, revision, charm_data)
        returnValue(charm_state)


class CharmState(object):
    """State of a charm registered in an environment."""

    def __init__(self, client, namespace, name, revision,
                 charm_data):
        self._client = client
        self._namespace = namespace
        self._name = name
        self._revision = revision

        self._metadata = MetaData()
        self._metadata.parse_serialization_data(charm_data["metadata"])

        self._config = ConfigOptions()
        self._config.parse(charm_data["config"])

        # Just a health check:
        assert self._metadata.name == name
        assert self._metadata.revision == revision

        self._sha256 = charm_data["sha256"]

        self._bundle_url = charm_data.get("url")

    @property
    def namespace(self):
        """The charm namespace."""
        return self._namespace

    @property
    def name(self):
        """The charm name."""
        return self._name

    @property
    def revision(self):
        """The monotonically increasing charm revision number.
        """
        return self._revision

    @property
    def bundle_url(self):
        """The url to the charm bundle in the provider storage."""
        return self._bundle_url

    @property
    def id(self):
        """The charm identifier, of the form <namespace>:<name>-<revision>.
        """
        return "%s:%s-%s" % \
               (self._namespace, self._name, self._revision)

    def get_metadata(self):
        """Return deferred MetaData."""
        return succeed(self._metadata)

    def get_config(self):
        """Return deferred ConfigOptions."""
        return succeed(self._config)

    def get_sha256(self):
        """Return deferred sha256 for the charm."""
        return succeed(self._sha256)
