"""Various unit placement strategies for use in deploy and add-unit.

The API used by the `place_unit` method is:

machine_state = yield placement_strategy(zk_client,
                                         machine_state_manager,
                                         unit_state)

The placement strategy is passed the machine manager for the
deployment and the unit_state it is attempting to place. According to
its policy it should yield back the machine_state for where it placed
the unit.
"""

from twisted.internet.defer import inlineCallbacks, returnValue

from juju.state.errors import NoUnusedMachines, InvalidPlacementPolicy

LOCAL_POLICY = "local"
UNASSIGNED_POLICY = "unassigned"


@inlineCallbacks
def _local_placement(client, machine_state_manager, unit_state):
    machine = yield machine_state_manager.get_machine_state(0)
    yield unit_state.assign_to_machine(machine)
    returnValue(machine)


@inlineCallbacks
def _unassigned_placement(client, machine_state_manager, unit_state):
    try:
        machine_state = yield unit_state.assign_to_unused_machine()
    except NoUnusedMachines:
        machine_state = yield machine_state_manager.add_machine_state()
        yield unit_state.assign_to_unused_machine()

    returnValue(machine_state)


_PLACEMENT_LOOKUP = {
    LOCAL_POLICY: _local_placement,
    UNASSIGNED_POLICY: _unassigned_placement,
    }


def place_unit(client, policy_name, machine_state_manager, unit_state):
    """Return machine state of unit_states assignment."""

    # default policy handling
    if policy_name is None:
        placement = _unassigned_placement
    else:
        placement = _PLACEMENT_LOOKUP.get(policy_name)
        if placement is None:
            raise InvalidPlacementPolicy(policy_name)

    return placement(client, machine_state_manager, unit_state)
